@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <gtk/gtk.h>
+#include <gdk/gdk.h>
 #include <gdk/gdkkeysyms.h>
 
 #include <hangul.h>
@@ -163,20 +164,25 @@ static void	im_hangul_preedit_normal     (GtkIMContextHangul *hic,
 					      gint start, gint end);
 
 static char*    im_hangul_get_candidate_string(GtkIMContextHangul *ic);
+static gboolean im_hangul_on_button_press    (GtkWidget *widget,
+					      GdkEvent *event,
+					      gpointer data);
 
 static void     im_hangul_ic_show_status_window     (GtkIMContextHangul *hcontext);
 static void     im_hangul_ic_hide_status_window     (GtkIMContextHangul *hcontext);
+static void     im_hangul_ic_update_status_window_position(GtkIMContextHangul *hic);
 static int      im_hangul_ic_get_toplevel_input_mode(GtkIMContextHangul *hcontext);
 static void     im_hangul_ic_set_toplevel_input_mode(GtkIMContextHangul *hcontext,
 						  int mode);
 
 static Toplevel*  toplevel_new(GtkWidget *toplevel_widget);
-static Toplevel*  toplevel_get(GdkWindow *window);
+static Toplevel*  toplevel_get_from_widget(GtkWidget *widget);
 static void       toplevel_append_context(Toplevel *toplevel,
 					  GtkIMContextHangul *context);
 static void       toplevel_remove_context(Toplevel *toplevel,
 					  GtkIMContextHangul *context);
 static void       toplevel_delete(Toplevel *toplevel);
+
 static GtkWidget* status_window_new(GtkWidget *parent);
 
 static void popup_candidate_window  (GtkIMContextHangul *hcontext);
@@ -627,6 +633,7 @@ im_hangul_ic_init (GtkIMContextHangul *hcontext)
 
   hcontext->client_window = NULL;
   hcontext->toplevel = NULL;
+  hcontext->button_press_handler = 0;
   hcontext->cursor.x = 0;
   hcontext->cursor.y = 0;
   hcontext->cursor.width = -1;
@@ -669,28 +676,43 @@ static void
 im_hangul_ic_set_client_window (GtkIMContext *context,
 			     GdkWindow *client_window)
 {
-  GtkIMContextHangul *hcontext;
+    GtkWidget *widget = NULL;
+    GtkIMContextHangul *hcontext;
 
-  g_return_if_fail (context != NULL);
-  g_return_if_fail (GTK_IS_IM_CONTEXT_HANGUL (context));
+    g_return_if_fail (context != NULL);
+    g_return_if_fail (GTK_IS_IM_CONTEXT_HANGUL (context));
 
-  hcontext = GTK_IM_CONTEXT_HANGUL(context);
+    hcontext = GTK_IM_CONTEXT_HANGUL(context);
 
-  if (hcontext->client_window == client_window)
-    return;
+    if (hcontext->client_window == client_window)
+	return;
 
-  if (hcontext->toplevel != NULL)
-    toplevel_remove_context(hcontext->toplevel, hcontext);
+    if (hcontext->toplevel != NULL)
+	toplevel_remove_context(hcontext->toplevel, hcontext);
 
-  if (client_window == NULL) {
-    hcontext->client_window = NULL;
-    hcontext->toplevel = NULL;
-    return;
-  }
+    if (client_window == NULL) {
+	gdk_window_get_user_data (hcontext->client_window, (gpointer)&widget);
+	if (widget != NULL) {
+	    g_signal_handler_disconnect (widget,
+		    hcontext->button_press_handler);
+	}
+	hcontext->button_press_handler = 0;
+	hcontext->client_window = NULL;
+	hcontext->toplevel = NULL;
+	return;
+    }
 
-  hcontext->client_window = client_window;
-  hcontext->toplevel = toplevel_get (client_window);
-  toplevel_append_context(hcontext->toplevel, hcontext);
+    hcontext->client_window = client_window;
+
+    gdk_window_get_user_data (hcontext->client_window, (gpointer)&widget);
+    if (widget != NULL) {
+	hcontext->toplevel = toplevel_get_from_widget (widget);
+	toplevel_append_context(hcontext->toplevel, hcontext);
+
+	hcontext->button_press_handler =
+		g_signal_connect (G_OBJECT (widget), "button-press-event",
+			G_CALLBACK(im_hangul_on_button_press), hcontext);
+    }
 }
 
 GtkIMContext *
@@ -726,7 +748,11 @@ static void
 im_hangul_set_input_mode_info (GdkWindow *window, int state)
 {
   if (window != NULL) {
+#if GTK_CHECK_VERSION(2, 24, 0)
+    GdkScreen *screen = gdk_window_get_screen(window);
+#else
     GdkScreen *screen = gdk_drawable_get_screen(window);
+#endif
     im_hangul_set_input_mode_info_for_screen (screen, state);
   }
 }
@@ -1003,6 +1029,8 @@ im_hangul_ic_cursor_location (GtkIMContext *context, GdkRectangle *area)
 
   hcontext = GTK_IM_CONTEXT_HANGUL(context);
   hcontext->cursor = *area;
+
+  im_hangul_ic_update_status_window_position(hcontext);
 }
 
 static inline gboolean
@@ -1566,90 +1594,63 @@ im_hangul_ic_filter_keypress (GtkIMContext *context, GdkEventKey *key)
 
 /* status window */
 static gboolean
-status_window_expose_event (GtkWidget *widget, GdkEventExpose *event)
+status_window_on_expose_event (GtkWidget *widget, GdkEventExpose *event)
 {
-  gdk_draw_rectangle (widget->window,
-		      widget->style->fg_gc[GTK_STATE_NORMAL],
-		      FALSE,
-		      0, 0,
-		      widget->allocation.width-1, widget->allocation.height-1);
-
-  return FALSE;
-}
-
-static gboolean
-status_window_configure	(GtkWidget *widget,
-			 GdkEventConfigure *event,
-			 Toplevel *toplevel)
-{
-  GdkRectangle rect;
-  GtkRequisition requisition;
-  gint y;
-
-  if (toplevel == NULL || toplevel->status == NULL)
+    gtk_paint_flat_box(widget->style, widget->window, GTK_STATE_NORMAL,
+	    GTK_SHADOW_OUT, NULL, widget, "tooltip", 0, 0,
+	    widget->allocation.width, widget->allocation.height);
     return FALSE;
-
-  gdk_window_get_frame_extents (widget->window, &rect);
-  gtk_widget_size_request (toplevel->status, &requisition);
-
-  if (rect.y + rect.height + requisition.height < gdk_screen_height ())
-    y = rect.y + rect.height;
-  else
-    y = gdk_screen_height () - requisition.height;
-
-  gtk_window_move (GTK_WINDOW(toplevel->status), rect.x, y);
-  return FALSE;
 }
 
 static GtkWidget*
 status_window_new(GtkWidget *parent)
 {
-  GtkWidget *window;
-  GtkWidget *frame;
-  GtkWidget *label;
+    GtkWidget *window;
+    GtkWidget *alignment;
+    GtkWidget *label;
 
-  if (parent == NULL)
-    return NULL;
+    if (parent == NULL)
+	return NULL;
 
-  window = gtk_window_new (GTK_WINDOW_POPUP);
+    window = gtk_window_new (GTK_WINDOW_POPUP);
 
-  gtk_container_set_border_width (GTK_CONTAINER(window), 1);
-  /* gtk_window_set_decorated (GTK_WINDOW(window), FALSE); */
-  gtk_widget_set_name (window, "imhangul_status");
-  gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
-  gtk_widget_set_app_paintable (window, TRUE);
+    gtk_container_set_border_width (GTK_CONTAINER(window), 1);
+    gtk_window_set_type_hint (GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_TOOLTIP);
 
-  frame = gtk_frame_new (NULL);
-  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_OUT);
-  gtk_widget_show (frame);
-  gtk_container_add (GTK_CONTAINER(window), frame);
+    gtk_widget_set_name (window, "imhangul_status");
+    gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
+    gtk_widget_set_app_paintable (window, TRUE);
 
-  /* hangul status window label */
-  label = gtk_label_new (_("hangul")); 
-  gtk_container_add (GTK_CONTAINER(frame), label);
-  gtk_widget_show (label);
+    alignment = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
+    gtk_alignment_set_padding (GTK_ALIGNMENT(alignment), 2, 2, 5, 5);
+    gtk_container_add (GTK_CONTAINER(window), alignment);
+    gtk_widget_show (alignment);
 
-  g_signal_connect (G_OBJECT(window), "expose-event",
-		   G_CALLBACK(status_window_expose_event), NULL);
+    /* hangul status window label */
+    label = gtk_label_new (_("hangul"));
+    gtk_container_add (GTK_CONTAINER(alignment), label);
+    gtk_widget_show (label);
 
-  return window;
+    g_signal_connect (G_OBJECT(window), "expose-event",
+	    G_CALLBACK(status_window_on_expose_event), NULL);
+
+    return window;
 }
 
 static void
 im_hangul_ic_show_status_window (GtkIMContextHangul *hcontext)
 {
-  g_return_if_fail (hcontext != NULL);
+    g_return_if_fail (hcontext != NULL);
 
-  if (pref_use_status_window && hcontext->toplevel != NULL) {
-    if (hcontext->toplevel->status == NULL) {
-      hcontext->toplevel->status =
-	status_window_new(hcontext->toplevel->widget);
-	status_window_configure (hcontext->toplevel->widget,
-				 NULL,
-				 hcontext->toplevel);
+    if (pref_use_status_window && hcontext->toplevel != NULL) {
+	if (hcontext->toplevel->status == NULL) {
+	    hcontext->toplevel->status =
+		status_window_new(hcontext->toplevel->widget);
+	}
+
+	im_hangul_ic_update_status_window_position(hcontext);
+	gtk_widget_show (hcontext->toplevel->status);
     }
-    gtk_widget_show (hcontext->toplevel->status);
-  }
 }
 
 static void
@@ -1662,21 +1663,39 @@ im_hangul_ic_hide_status_window (GtkIMContextHangul *hcontext)
   }
 }
 
-static GtkWidget *
-get_toplevel_widget (GdkWindow *window)
+static void
+im_hangul_ic_update_status_window_position (GtkIMContextHangul* hic)
 {
-  GtkWidget *gtk_toplevel;
-  gpointer ptr;
+    int x = 0;
+    int y = 0;
 
-  if (window == NULL)
-    return NULL;
+    if (hic == NULL)
+	return;
 
-  gdk_window_get_user_data (window, &ptr);
-  memcpy(&gtk_toplevel, &ptr, sizeof(gtk_toplevel));
-  if (gtk_toplevel != NULL)
-    gtk_toplevel = gtk_widget_get_toplevel(GTK_WIDGET(gtk_toplevel));
+    if (hic->client_window == NULL)
+	return;
 
-  return gtk_toplevel;
+    if (hic->toplevel == NULL || hic->toplevel->status == NULL)
+	return;
+
+    gdk_window_get_origin (hic->client_window, &x, &y);
+
+    if (hic->cursor.x < 0) {
+	/* show status window below client window
+	 * if the cursor position is not updated */
+	int height = 0;
+#if GTK_CHECK_VERSION(2, 24, 0)
+	height = gdk_window_get_height(hic->client_window);
+#else
+	gdk_drawable_get_size(hic->client_window, NULL, &height);
+#endif
+	y += height + 3;
+    } else {
+	x += hic->cursor.x + 3;
+	y += hic->cursor.y + hic->cursor.height + 3;
+    }
+
+    gtk_window_move (GTK_WINDOW(hic->toplevel->status), x, y);
 }
 
 static void
@@ -1688,6 +1707,18 @@ toplevel_destroy(Toplevel *toplevel)
   }
 }
 
+static gboolean
+toplevel_on_configure_event(GtkWidget *widget,
+			 GdkEventConfigure *event,
+			 Toplevel *toplevel)
+{
+    if (current_focused_ic != NULL) {
+	GtkIMContextHangul* hic = GTK_IM_CONTEXT_HANGUL(current_focused_ic);
+	im_hangul_ic_update_status_window_position (hic);
+    }
+    return FALSE;
+}
+
 static Toplevel *
 toplevel_new(GtkWidget *toplevel_widget)
 {
@@ -1703,8 +1734,7 @@ toplevel_new(GtkWidget *toplevel_widget)
 			     G_CALLBACK(toplevel_destroy), toplevel);
   toplevel->configure_handler_id = 
 	    g_signal_connect (G_OBJECT(toplevel->widget), "configure-event",
-			     G_CALLBACK(status_window_configure),
-			     toplevel);
+			     G_CALLBACK(toplevel_on_configure_event), NULL);
 
   g_object_set_data(G_OBJECT(toplevel_widget),
 		     "gtk-imhangul-toplevel-info", toplevel);
@@ -1712,12 +1742,12 @@ toplevel_new(GtkWidget *toplevel_widget)
 }
 
 static Toplevel *
-toplevel_get(GdkWindow *window)
+toplevel_get_from_widget(GtkWidget *widget)
 {
   Toplevel *toplevel = NULL;
   GtkWidget *toplevel_widget;
 
-  toplevel_widget = get_toplevel_widget (window);
+  toplevel_widget = gtk_widget_get_toplevel (widget);
   if (toplevel_widget == NULL) {
     return NULL;
   }
@@ -1791,6 +1821,13 @@ im_hangul_ic_set_toplevel_input_mode(GtkIMContextHangul *hcontext, int mode)
     hcontext->toplevel->input_mode = mode;
 }
 
+static gboolean
+im_hangul_on_button_press(GtkWidget *widget, GdkEvent *event, gpointer data)
+{
+    im_hangul_ic_reset(data);
+    return false;
+}
+
 /*
  * candidate selection window
  */
@@ -1925,8 +1962,8 @@ im_hangul_init(void)
   }
 
   if (hanja_keys->len == 0) {
-    im_hangul_accel_list_append(hangul_keys, GDK_Hangul_Hanja, 0);
-    im_hangul_accel_list_append(hangul_keys, GDK_F9, 0);
+    im_hangul_accel_list_append(hanja_keys, GDK_Hangul_Hanja, 0);
+    im_hangul_accel_list_append(hanja_keys, GDK_F9, 0);
   }
 
   /* install gtk key snooper
@@ -2102,14 +2139,17 @@ candidate_on_expose (GtkWidget *widget,
 		     GdkEventExpose *event,
 		     gpointer data)
 {
-  GtkStyle *style;
-  GtkAllocation alloc;
-
-  style = gtk_widget_get_style(widget);
-  alloc = GTK_WIDGET(widget)->allocation;
-  gdk_draw_rectangle(widget->window, style->black_gc,
-		     FALSE,
-		     0, 0, alloc.width - 1, alloc.height - 1);
+#if GTK_CHECK_VERSION(2, 8, 0)
+    cairo_t* cr = gdk_cairo_create(widget->window);
+    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
+    cairo_rectangle(cr, 0, 0,
+	    widget->allocation.width, widget->allocation.height);
+    cairo_stroke(cr);
+    cairo_destroy(cr);
+#else
+    gdk_draw_rectangle(widget->window, widget->style->black_gc, FALSE, 0, 0,
+	    widget->allocation.width - 1, widget->allocation.height - 1);
+#endif
 }
 
 static void
@@ -2129,7 +2169,7 @@ candidate_update_cursor(Candidate *candidate)
 static void
 candidate_set_window_position (Candidate *candidate)
 {
-    gint width = 0, height = 0;
+    gint height = 0;
     gint absx = 0, absy = 0;
     gint root_w, root_h, cand_w, cand_h;
     GtkRequisition requisition;
@@ -2138,7 +2178,11 @@ candidate_set_window_position (Candidate *candidate)
       return;
 
     gdk_window_get_origin (GDK_WINDOW(candidate->parent), &absx, &absy);
-    gdk_drawable_get_size (GDK_DRAWABLE(candidate->parent), &width, &height);
+#if GTK_CHECK_VERSION(2, 24, 0)
+    height = gdk_window_get_height(GDK_WINDOW(candidate->parent));
+#else
+    gdk_drawable_get_size (GDK_DRAWABLE(candidate->parent), NULL, &height);
+#endif
 
     root_w = gdk_screen_width();
     root_h = gdk_screen_height();
@@ -2209,6 +2253,8 @@ candidate_create_window(Candidate *candidate)
   GtkCellRenderer *renderer;
 
   candidate->window = gtk_window_new(GTK_WINDOW_POPUP);
+  gtk_window_set_type_hint (GTK_WINDOW(candidate->window), GDK_WINDOW_TYPE_HINT_TOOLTIP);
+  gtk_widget_set_app_paintable (candidate->window, TRUE);
 
   candidate_update_list(candidate);
 
